import pandas as pd
import yfinance as yf
import mplfinance as mpf
import matplotlib.pyplot as plt
from matplotlib.widgets import TextBox, RadioButtons


class DataFetcher:
    def __init__(self, asset: str, period: str = '1mo', interval: str = '15m'):
        self.asset = asset
        self.period = period
        self.interval = interval

    def fetch_data(self) -> pd.DataFrame:
        data = yf.download(self.asset, period=self.period, interval=self.interval)
        if data.empty:
            raise ValueError(f"No data fetched for ticker symbol '{self.asset}'. Check the ticker symbol.")
        return data

class DataManipulator:
    def __init__(self, df: pd.DataFrame):
        self.df = df

    def add_indicators(self) -> pd.DataFrame:
        df_with_bollinger = self.add_bollinger_bands(df=self.df)
        return df_with_bollinger

    @staticmethod
    def add_bollinger_bands(df: pd.DataFrame) -> pd.DataFrame:
        df['SMA'] = df['Close'].rolling(window=20).mean()
        df['STD'] = df['Close'].rolling(window=20).std().round(3)
        df['Upper_Bollinger_Band'] = df['SMA'] + (df['STD'] * 2)
        df['Lower_Bollinger_Band'] = df['SMA'] - (df['STD'] * 2)
        return df

class ChartRenderer:

    def __init__(
            self,
            asset: str,
            df: pd.DataFrame
    ):
        self.asset = asset
        self.df = df
        self.current_plot_type = 'candle'
        self.current_period = '1mo'
        self.current_interval = '15m'
        self.indicators_width = 0.35
        self.ax = None

    def plot_data(self) -> None:
        df = self.df.copy()
        df.index.name = 'Date'

        add_plots = [
            mpf.make_addplot(df['SMA'], color='green', label='SMA', width=self.indicators_width),
            mpf.make_addplot(df['Upper_Bollinger_Band'], color='red', linestyle='--', label='Upper Bollinger',
                             width=self.indicators_width),
            mpf.make_addplot(df['Lower_Bollinger_Band'], color='blue', linestyle='--', label='Lower Bollinger',
                             width=self.indicators_width)
        ]

        fig, axlist = mpf.plot(df, type=self.current_plot_type, style='charles', addplot=add_plots, ylabel='Price',
                               volume=False, figsize=(14, 7), returnfig=True)
        self.ax = axlist[0]
        self.ax.set_title(f'{self.asset} Price Chart')

        self.add_widget_box(fig)
        plt.show()

    def update_plot(self) -> None:
        self.ax.clear()
        self.ax.set_title(f'{self.asset} Price Chart')

        add_plots_updated = [
            mpf.make_addplot(self.df['SMA'], color='green', label='SMA', width=self.indicators_width, ax=self.ax),
            mpf.make_addplot(self.df['Upper_Bollinger_Band'], color='red', linestyle='--', label='Upper Bollinger',
                             width=self.indicators_width, ax=self.ax),
            mpf.make_addplot(self.df['Lower_Bollinger_Band'], color='blue', linestyle='--', label='Lower Bollinger',
                             width=self.indicators_width, ax=self.ax)
        ]

        mpf.plot(self.df, type=self.current_plot_type, style='charles', addplot=add_plots_updated, ax=self.ax,
                 volume=False)

        plt.draw()

    def add_widget_box(self, fig) -> None:
        ax_radio_plot_type = plt.axes([0.05, 0.85, 0.1, 0.1])
        radio_plot_type = RadioButtons(ax_radio_plot_type, ['candle', 'line', 'ohlc'])
        radio_plot_type.on_clicked(self.change_plot_type)

        ax_radio_period = plt.axes([0.05, 0.6, 0.1, 0.25])
        radio_period = RadioButtons(ax_radio_period,
                                    ['1d', '5d', '1mo', '3mo', '6mo', '1y', '2y', '5y', '10y', 'ytd', 'max'],
                                    active=2)
        radio_period.on_clicked(self.change_period)

        ax_radio_interval = plt.axes([0.05, 0.3, 0.1, 0.25])
        radio_interval = RadioButtons(ax_radio_interval,
                                      ['1m', '2m', '5m', '15m', '30m', '60m', '90m', '1h', '1d', '5d', '1wk',
                                       '1mo', '3mo'], active=3)
        radio_interval.on_clicked(self.change_interval)

        ax_textbox = plt.axes([0.05, 0.10, 0.1, 0.05])
        textbox = TextBox(ax_textbox, 'Input', initial=self.asset)
        textbox.on_submit(self.submit_text)

        plt.show()

    def change_plot_type(self, label: str) -> None:
        self.current_plot_type = label
        self.update_plot()

    def change_period(self, label: str) -> None:
        self.current_period = label
        self.update_asset_data()

    def change_interval(self, label: str) -> None:
        self.current_interval = label
        self.update_asset_data()

    def submit_text(self, text: str) -> None:
        self.asset = text
        self.update_asset_data()

    def update_asset_data(self) -> None:
        fetcher = DataFetcher(self.asset, self.current_period, self.current_interval)
        self.df = fetcher.fetch_data()
        manipulator = DataManipulator(self.df)
        self.df = manipulator.add_indicators()
        self.update_plot()


class Stonks:
    def __init__(self, asset: str):
        self.asset = asset
        self.data_fetcher = DataFetcher(asset)
        self.data_frame = self.data_fetcher.fetch_data()
        self.data_manipulator = DataManipulator(self.data_frame)
        self.data_frame = self.data_manipulator.add_indicators()
        self.chart_renderer = ChartRenderer(asset, self.data_frame)

    def run(self) -> None:
        self.chart_renderer.plot_data()

def run_stonks_analysis(asset: str) -> None:
    stonks = Stonks(asset)
    stonks.run()



from io import StringIO
import pytest
import pandas as pd
from unittest.mock import patch, MagicMock

from stonks.stonks import DataFetcher, DataManipulator, ChartRenderer


class TestDataFetcher:

    def test_given_empty_dataframe_is_returned_when_passing_dataframe_then_exception_is_being_thrown(self):
        # given
        asset = 'INVALID_ASSET'
        # when
        with patch('stonks.stonks.yf.download', return_value=pd.DataFrame()):
            # then
            with pytest.raises(ValueError):
                fetcher = DataFetcher(asset)
                fetcher.fetch_data()

    def test_given_incorrect_asset_name_when_downloading_historical_data_then_empty_dataframe_is_returned(self):
        # given
        asset = 'NOT_STONKS'
        # when
        with patch('stonks.stonks.DataFetcher.fetch_data', return_value=pd.DataFrame()):
            fetcher = DataFetcher(asset)
            df = fetcher.fetch_data()
        # then
        assert df.empty, "Expected empty dataframe for invalid asset"

    def test_given_correct_asset_name_when_downloading_historical_data_then_non_empty_dataframe_is_returned(self):
        # given
        asset = 'EURUSD=X'
        fetcher = DataFetcher(asset)
        # when
        df = fetcher.fetch_data()

        # then
        assert not df.empty, "Expected non-empty dataframe for valid asset"
        expected_columns = ['Open', 'High', 'Low', 'Close', 'Volume']
        assert all(column in df.columns for column in expected_columns), "Expected columns missing from DataFrame"


class TestChartRenderer:

    def test_change_plot_type_updates_plot(self):
        # given
        asset = 'EURUSD=X'
        df = pd.DataFrame({
            'Close': [10, 20, 30],
            'SMA': [15, 25, 35],
            'Upper_Bollinger_Band': [20, 30, 40],
            'Lower_Bollinger_Band': [10, 20, 30]
        })
        chart_renderer = ChartRenderer(asset, df)
        chart_renderer.update_plot = MagicMock()

        # when
        chart_renderer.change_plot_type('line')

        # then
        chart_renderer.update_plot.assert_called_once()
        assert chart_renderer.current_plot_type == 'line'

    def test_change_period_triggers_update_asset_data(self):
        # given
        asset = 'EURUSD=X'
        df = pd.DataFrame({
            'Close': [10, 20, 30],
            'SMA': [15, 25, 35],
            'Upper_Bollinger_Band': [20, 30, 40],
            'Lower_Bollinger_Band': [10, 20, 30]
        })
        chart_renderer = ChartRenderer(asset, df)
        chart_renderer.update_asset_data = MagicMock()

        # when
        chart_renderer.change_period('3mo')

        # then
        chart_renderer.update_asset_data.assert_called_once()
        assert chart_renderer.current_period == '3mo'

    def test_change_interval_triggers_update_asset_data(self):
        # given
        asset = 'EURUSD=X'
        df = pd.DataFrame({
            'Close': [10, 20, 30],
            'SMA': [15, 25, 35],
            'Upper_Bollinger_Band': [20, 30, 40],
            'Lower_Bollinger_Band': [10, 20, 30]
        })
        chart_renderer = ChartRenderer(asset, df)
        chart_renderer.update_asset_data = MagicMock()

        # when
        chart_renderer.change_interval('1d')

        # then
        chart_renderer.update_asset_data.assert_called_once()
        assert chart_renderer.current_interval == '1d'

    def test_submit_text_triggers_update_asset_data(self):
        # given
        asset = 'AAPL'
        df = pd.DataFrame({
            'Close': [10, 20, 30],
            'SMA': [15, 25, 35],
            'Upper_Bollinger_Band': [20, 30, 40],
            'Lower_Bollinger_Band': [10, 20, 30]
        })
        chart_renderer = ChartRenderer(asset, df)
        chart_renderer.update_asset_data = MagicMock()  # Mock update_asset_data method

        # when
        chart_renderer.submit_text('GOOGL')

        # then
        chart_renderer.update_asset_data.assert_called_once()
        assert chart_renderer.asset == 'GOOGL'


class TestDataManipulator:

    def test_given_ohlc_historical_dataframe_when_calculating_indicators_then_is_dataframe_with_correct_data_returned(self):
        # given
        data = {
            'Close': [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 29, 28, 27, 26]
        }
        df = pd.DataFrame(data)
        manipulator = DataManipulator(df)

        # when
        df_with_indicators = manipulator.add_indicators()

        # then
        expected_csv = StringIO(
            """Close,SMA,STD,Upper_Bollinger_Band,Lower_Bollinger_Band
            10,,,,
            11,,,,
            12,,,,
            13,,,,
            14,,,,
            15,,,,
            16,,,,
            17,,,,
            18,,,,
            19,,,,
            20,,,,
            21,,,,
            22,,,,
            23,,,,
            24,,,,
            25,,,,
            26,,,,
            27,,,,
            28,,,,
            29,19.5,5.916,31.332,7.668
            30,20.5,5.916,32.332,8.668
            29,21.4,5.762,32.924,9.876
            28,22.2,5.493,33.186,11.214
            27,22.9,5.139,33.178,12.622
            26,23.5,4.730,32.960,14.040
            """
        )
        expected_df = pd.read_csv(expected_csv)

        pd.testing.assert_frame_equal(df_with_indicators.reset_index(drop=True), expected_df.reset_index(drop=True))

    def test_bollinger_bands_with_short_history(self):
        # given
        data = {'Close': [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]}
        df = pd.DataFrame(data)
        manipulator = DataManipulator(df)

        # when
        df_with_bollinger = manipulator.add_indicators()

        # then
        assert 'SMA' in df_with_bollinger.columns
        assert 'Upper_Bollinger_Band' in df_with_bollinger.columns
        assert 'Lower_Bollinger_Band' in df_with_bollinger.columns

        assert df_with_bollinger['SMA'].isna().all(), "SMA should be NaN for less than 20 data points"
        assert df_with_bollinger['Upper_Bollinger_Band'].isna().all(), "Upper Bollinger Band should be NaN for less than 20 data points"
        assert df_with_bollinger['Lower_Bollinger_Band'].isna().all(), "Lower Bollinger Band should be NaN for less than 20 data points"



from stonks.stonks import run_stonks_analysis
from stonks.version import version

__version__ = version.version
__author__ = "Daniel Lasota"
__description__ = "Script that scrapes market Data from Yahoo Finance Api and performs base analysis with plotting"
__email__ = "grossmann.root@gmail.com"
__url__ = "https://www.youtube.com/watch?v=dQw4w9WgXcQ"



from setuptools import setup, find_packages

setup(
    name="stonks",
    version="0.1.2",
    packages=find_packages(),
    install_requires=[
        'pytest',
        'pandas',
        'yfinance',
        'matplotlib',
        'mplfinance',
        'matplotlib'
    ],
    entry_points={
        'console_scripts': [
            'run_stonks_analysis=stonks.stonks:run_stonks_analysis',
        ],
    },
    author="Daniel Lasota",
    author_email="grossmann.root@gmail.com",
    description="stonks stonks",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/DanielLasota/Stonks",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>3.8, <3.9',
)